/**
 *  Classes.h
 *
 *  List of all declared classes
 *
 *  @copyright 2014 - 2017 Copernica BV
 */

/**
 *  Include guard
 */
#pragma once

/**
 *  Set up namespace
 */
namespace AMQP {

/**
 *  All classes defined by this library
 */
class Array;
class BasicDeliverFrame;
class BasicGetOKFrame;
class BasicHeaderFrame;
class BasicReturnFrame;
class BasicAckFrame;
class BasicNackFrame;
class BodyFrame;
class Channel;
class Connection;
class ConnectionHandler;
class ConnectionImpl;
class CopiedBuffer;
class Exchange;
class Frame;
class Login;
class Monitor;
class OutBuffer;
class InBuffer;
class Table;

/**
 *  End of namespace
 */
}

